@@ -17,13 +17,15 @@ from sklearn.metrics.pairwise import cosine_similarity
 
 # BERTopic
 from .ctfidf import ClassTFIDF
-from .utils import create_logger, check_documents_type
+from .utils import create_logger, check_documents_type, check_embeddings_shape
 logger = create_logger()
 
 
 class BERTopic:
     """
-    Transformer-based model for Topic Modeling
+    BERTopic is a topic modeling technique that leverages BERT embeddings and
+    c-TF-IDF to create dense clusters allowing for easily interpretable topics
+    whilst keeping important words in the topic descriptions.
 
     Arguments:
         bert_model: Model to use. Overview of options can be found here
@@ -32,40 +34,68 @@ class BERTopic:
         nr_topics: Specifying the number of topics will reduce the initial
                    number of topics to the value specified. This reduction can take
                    a while as each reduction in topics (-1) activates a c-TF-IDF calculation.
-                   IF this is set to None, no reduction is applied.
+                   IF this is set to None, no reduction is applied. Use "auto" to automatically
+                   reduce topics that have a similarity of at least 0.9, do not maps all others.
         n_gram_range: The n-gram range for the CountVectorizer.
                       Advised to keep high values between 1 and 3.
                       More would likely lead to memory issues.
+                      Note that this will not be used if you pass in your own CountVectorizer.
         min_topic_size: The minimum size of the topic.
         n_neighbors: The size of local neighborhood (in terms of number of neighboring sample points) used
                      for manifold approximation (UMAP).
         n_components: The dimension of the space to embed into when reducing dimensionality with UMAP.
         stop_words: Stopwords that can be used as either a list of strings, or the name of the
                     language as a string. For example: 'english' or ['the', 'and', 'I'].
+                    Note that this will not be used if you pass in your own CountVectorizer.
         verbose: Changes the verbosity of the model, Set to True if you want
                  to track the stages of the model.
+        vectorizer: Pass in your own CountVectorizer from scikit-learn
 
     Usage:
-        ```python
-        from bertopic import BERTopic
-        from sklearn.datasets import fetch_20newsgroups
 
-        docs = fetch_20newsgroups(subset='all')['data']
+    ```python
+    from bertopic import BERTopic
+    from sklearn.datasets import fetch_20newsgroups
 
-        model = BERTopic("distilbert-base-nli-mean-tokens", verbose=True)
-        topics = model.fit_transform(docs)
-        ```
+    docs = fetch_20newsgroups(subset='all')['data']
+
+    model = BERTopic("distilbert-base-nli-mean-tokens", verbose=True)
+    topics = model.fit_transform(docs)
+    ```
+
+    If you want to use your own embeddings, use it as follows:
+
+    ```python
+    from bertopic import BERTopic
+    from sklearn.datasets import fetch_20newsgroups
+    from sentence_transformers import SentenceTransformer
+
+    # Create embeddings
+    docs = fetch_20newsgroups(subset='all')['data']
+    sentence_model = SentenceTransformer("distilbert-base-nli-mean-tokens")
+    embeddings = sentence_model.encode(docs, show_progress_bar=True)
+
+    # Create topic model
+    model = BERTopic(None, verbose=True)
+    topics = model.fit_transform(docs, embeddings)
+    ```
+
+    Due to the stochastisch nature of UMAP, the results from BERTopic might differ
+    and the quality can degrade. Using your own embeddings allows you to
+    try out BERTopic several times until you find the topics that suit
+    you best.
     """
     def __init__(self,
                  bert_model: str = 'distilbert-base-nli-mean-tokens',
                  top_n_words: int = 20,
-                 nr_topics: int = None,
+                 nr_topics: Union[int, str] = None,
                  n_gram_range: Tuple[int, int] = (1, 1),
                  min_topic_size: int = 30,
                  n_neighbors: int = 15,
                  n_components: int = 5,
                  stop_words: Union[str, List[str]] = None,
-                 verbose: bool = False):
+                 verbose: bool = False,
+                 vectorizer: CountVectorizer = None):
         self.bert_model = bert_model
         self.top_n_words = top_n_words
         self.nr_topics = nr_topics
@@ -74,6 +104,7 @@ class BERTopic:
         self.n_neighbors = n_neighbors
         self.n_components = n_components
         self.stop_words = stop_words
+        self.vectorizer = vectorizer or CountVectorizer(ngram_range=self.n_gram_range, stop_words=self.stop_words)
 
         self.umap_model = None
         self.cluster_model = None
@@ -87,27 +118,89 @@ class BERTopic:
         else:
             logger.setLevel(logging.WARNING)
 
-    def fit(self, documents: List[str]):
+    def fit(self,
+            documents: List[str],
+            embeddings: np.ndarray = None):
         """ Fit the models (Bert, UMAP, and, HDBSCAN) on a collection of documents and generate topics
 
         Arguments:
             documents: A list of documents to fit on
+            embeddings: Pre-trained document embeddings. These can be used
+                        instead of the sentence-transformer model
+
+        Usage:
+
+        ```python
+        from bertopic import BERTopic
+        from sklearn.datasets import fetch_20newsgroups
+
+        docs = fetch_20newsgroups(subset='all')['data']
+        model = BERTopic("distilbert-base-nli-mean-tokens", verbose=True).fit(docs)
+        ```
+
+        If you want to use your own embeddings, use it as follows:
+
+        ```python
+        from bertopic import BERTopic
+        from sklearn.datasets import fetch_20newsgroups
+        from sentence_transformers import SentenceTransformer
+
+        # Create embeddings
+        docs = fetch_20newsgroups(subset='all')['data']
+        sentence_model = SentenceTransformer("distilbert-base-nli-mean-tokens")
+        embeddings = sentence_model.encode(docs, show_progress_bar=True)
+
+        # Create topic model
+        model = BERTopic(None, verbose=True).fit(docs, embeddings)
+        ```
         """
         check_documents_type(documents)
-        self.fit_transform(documents)
+        self.fit_transform(documents, embeddings)
         return self
 
     def fit_transform(self,
-                      documents: List[str]) -> Tuple[List[int],
-                                                     np.ndarray]:
+                      documents: List[str],
+                      embeddings: np.ndarray = None) -> Tuple[List[int],
+                                                              np.ndarray]:
         """ Fit the models on a collection of documents, generate topics, and return the docs with topics
 
         Arguments:
             documents: A list of documents to fit on
+            embeddings: Pre-trained document embeddings. These can be used
+                        instead of the sentence-transformer model
 
         Returns:
             predictions: Topic predictions for each documents
             probabilities: The topic probability distribution
+
+        Usage:
+
+        ```python
+        from bertopic import BERTopic
+        from sklearn.datasets import fetch_20newsgroups
+
+        docs = fetch_20newsgroups(subset='all')['data']
+
+        model = BERTopic("distilbert-base-nli-mean-tokens", verbose=True)
+        topics = model.fit_transform(docs)
+        ```
+
+        If you want to use your own embeddings, use it as follows:
+
+        ```python
+        from bertopic import BERTopic
+        from sklearn.datasets import fetch_20newsgroups
+        from sentence_transformers import SentenceTransformer
+
+        # Create embeddings
+        docs = fetch_20newsgroups(subset='all')['data']
+        sentence_model = SentenceTransformer("distilbert-base-nli-mean-tokens")
+        embeddings = sentence_model.encode(docs, show_progress_bar=True)
+
+        # Create topic model
+        model = BERTopic(None, verbose=True)
+        topics = model.fit_transform(docs, embeddings)
+        ```
         """
         check_documents_type(documents)
         documents = pd.DataFrame({"Document": documents,
@@ -115,7 +208,9 @@ class BERTopic:
                                   "Topic": None})
 
         # Extract BERT sentence embeddings
-        embeddings = self._extract_embeddings(documents.Document)
+        if not isinstance(embeddings, np.ndarray):
+            check_embeddings_shape(embeddings, documents)
+            embeddings = self._extract_embeddings(documents.Document)
 
         # Reduce dimensionality with UMAP
         umap_embeddings = self._reduce_dimensionality(embeddings)
@@ -134,20 +229,55 @@ class BERTopic:
 
         return predictions, probabilities
 
-    def transform(self, documents: Union[str, List[str]]) -> Tuple[List[int], np.ndarray]:
+    def transform(self,
+                  documents: Union[str, List[str]],
+                  embeddings: np.ndarray = None) -> Tuple[List[int], np.ndarray]:
         """ After having fit a model, use transform to predict new instances
 
         Arguments:
             documents: A single document or a list of documents to fit on
+            embeddings: Pre-trained document embeddings. These can be used
+                        instead of the sentence-transformer model.
 
         Returns:
             predictions: Topic predictions for each documents
             probabilities: The topic probability distribution
+
+        Usage:
+
+        ```python
+        from bertopic import BERTopic
+        from sklearn.datasets import fetch_20newsgroups
+
+        docs = fetch_20newsgroups(subset='all')['data']
+        model = BERTopic("distilbert-base-nli-mean-tokens", verbose=True).fit(docs)
+        topics = model.transform(docs)
+        ```
+
+        If you want to use your own embeddings:
+
+        ```python
+        from bertopic import BERTopic
+        from sklearn.datasets import fetch_20newsgroups
+        from sentence_transformers import SentenceTransformer
+
+        # Create embeddings
+        docs = fetch_20newsgroups(subset='all')['data']
+        sentence_model = SentenceTransformer("distilbert-base-nli-mean-tokens")
+        embeddings = sentence_model.encode(docs, show_progress_bar=True)
+
+        # Create topic model
+        model = BERTopic(None, verbose=True).fit(docs, embeddings)
+        topics = model.transform(docs, embeddings)
+        ```
         """
         if isinstance(documents, str):
             documents = [documents]
 
-        embeddings = self._extract_embeddings(documents)
+        if not isinstance(embeddings, np.ndarray):
+            check_embeddings_shape(embeddings, documents)
+            embeddings = self._extract_embeddings(documents)
+
         umap_embeddings = self.umap_model.transform(embeddings)
         probabilities = hdbscan.membership_vector(self.cluster_model, umap_embeddings)
         predictions, _ = hdbscan.approximate_predict(self.cluster_model, umap_embeddings)
@@ -262,7 +392,7 @@ class BERTopic:
             words: The names of the words to which values were given
         """
         documents = documents_per_topic.Document.values
-        count = CountVectorizer(ngram_range=self.n_gram_range, stop_words=self.stop_words).fit(documents)
+        count = self.vectorizer.fit(documents)
         words = count.get_feature_names()
         X = count.transform(documents)
         transformer = ClassTFIDF().fit(X, n_samples=m)
@@ -293,27 +423,77 @@ class BERTopic:
                        for i, label in enumerate(labels)}
 
     def get_topics(self) -> Dict[str, Tuple[str, float]]:
-        """ Return topics with top n words and their c-TF-IDF score """
+        """ Return topics with top n words and their c-TF-IDF score
+
+        Usage:
+
+        ```python
+        all_topics = model.get_topics()
+        ```
+        """
         return self.topics
 
     def get_topic(self, topic: int) -> Union[Dict[str, Tuple[str, float]], bool]:
-        """ Return top n words for a specific topic and their c-TF-IDF scores """
+        """ Return top n words for a specific topic and their c-TF-IDF scores
+
+        Usage:
+
+        ```python
+        topic = model.get_topic(12)
+        ```
+        """
         if self.topics.get(topic):
             return self.topics[topic]
         else:
             return False
 
     def get_topics_freq(self) -> pd.DataFrame:
-        """ Return the the size of topics (descending order) """
+        """ Return the the size of topics (descending order)
+
+        Usage:
+
+        ```python
+        frequency = model.get_topics_freq()
+        ```
+        """
         return pd.DataFrame(self.topic_sizes.items(), columns=['Topic', 'Count']).sort_values("Count", ascending=False)
 
     def get_topic_freq(self, topic: int) -> int:
-        """ Return the the size of a topic """
+        """ Return the the size of a topic
+
+        Arguments:
+             topic: the name of the topic as retrieved by get_topics
+
+        Usage:
+
+        ```python
+        frequency = model.get_topic_freq(12)
+        ```
+        """
         return self.topic_sizes.items()[topic]
 
     def _reduce_topics(self, documents: pd.DataFrame, c_tf_idf: np.ndarray) -> pd.DataFrame:
         """ Reduce topics to self.nr_topics
 
+        Arguments:
+            documents: Dataframe with documents and their corresponding IDs and Topics
+            c_tf_idf: c-TF-IDF matrix
+
+        Returns:
+            documents: Updated dataframe with documents and the reduced number of Topics
+        """
+        if isinstance(self.nr_topics, int):
+            documents = self._reduce_to_n_topics(c_tf_idf, documents)
+        elif isinstance(self.nr_topics, str):
+            documents = self._auto_reduce_topics(c_tf_idf, documents)
+        else:
+            raise ValueError("nr_topics needs to be an int or 'auto'! ")
+
+        return documents
+
+    def _reduce_to_n_topics(self, c_tf_idf, documents):
+        """ Reduce topics to self.nr_topics
+
         Arguments:
             documents: Dataframe with documents and their corresponding IDs and Topics
             c_tf_idf: c-TF-IDF matrix
@@ -323,16 +503,16 @@ class BERTopic:
         """
         self.mapped_topics = {}
         initial_nr_topics = len(self.get_topics())
-        nr_to_reduce = initial_nr_topics - self.nr_topics
 
-        for _ in range(nr_to_reduce):
-            # Create topic similarity matrix
-            similarities = cosine_similarity(c_tf_idf)
-            np.fill_diagonal(similarities, 0)
+        # Create topic similarity matrix
+        similarities = cosine_similarity(c_tf_idf)
+        np.fill_diagonal(similarities, 0)
 
+        while len(self.get_topics_freq()) > self.nr_topics + 1:
             # Find most similar topic to least common topic
             topic_to_merge = self.get_topics_freq().iloc[-1].Topic
             topic_to_merge_into = np.argmax(similarities[topic_to_merge + 1]) - 1
+            similarities[:, topic_to_merge + 1] = -1
 
             # Update Topic labels
             documents.loc[documents.Topic == topic_to_merge, "Topic"] = topic_to_merge_into
@@ -340,12 +520,57 @@ class BERTopic:
 
             # Update new topic content
             self._update_topic_size(documents)
-            self._extract_topics(documents, topic_reduction=True)
+
+        self._extract_topics(documents, topic_reduction=True)
 
         if initial_nr_topics <= self.nr_topics:
             logger.info(f"Since {initial_nr_topics} were found, they could not be reduced to {self.nr_topics}")
         else:
-            logger.info(f"Reduced number of topics from {initial_nr_topics} to {self.nr_topics}")
+            logger.info(f"Reduced number of topics from {initial_nr_topics} to {len(self.get_topics_freq())}")
+
+        return documents
+
+    def _auto_reduce_topics(self, c_tf_idf, documents):
+        """ Reduce the number of topics as long as it exceeds a minimum similarity of 0.9
+
+        Arguments:
+            documents: Dataframe with documents and their corresponding IDs and Topics
+            c_tf_idf: c-TF-IDF matrix
+
+        Returns:
+            documents: Updated dataframe with documents and the reduced number of Topics
+        """
+        initial_nr_topics = len(self.get_topics())
+
+        # Create topic similarity matrix
+        similarities = cosine_similarity(c_tf_idf)
+        np.fill_diagonal(similarities, 0)
+
+        # Do not map the top 10% most frequent topics
+        not_mapped = int(np.ceil(len(self.get_topics_freq()) * 0.1))
+        to_map = self.get_topics_freq().Topic.values[not_mapped:][::-1]
+
+        self.mapped_topics = {}
+        has_mapped = []
+
+        for topic_to_merge in to_map:
+            similarity = np.max(similarities[topic_to_merge + 1])
+            topic_to_merge_into = np.argmax(similarities[topic_to_merge + 1]) - 1
+
+            # Only map topics if they have a high similarity
+            if (similarity > 0.915) & (topic_to_merge_into not in has_mapped):
+                # Update Topic labels
+                documents.loc[documents.Topic == topic_to_merge, "Topic"] = topic_to_merge_into
+                self.mapped_topics[topic_to_merge] = topic_to_merge_into
+                similarities[:, topic_to_merge + 1] = -1
+
+                # Update new topic content
+                self._update_topic_size(documents)
+                has_mapped.append(topic_to_merge)
+
+        _ = self._extract_topics(documents, topic_reduction=True)
+
+        logger.info(f"Reduced number of topics from {initial_nr_topics} to {len(self.get_topics_freq())}")
 
         return documents
 
@@ -382,6 +607,17 @@ class BERTopic:
                              All others are ignored.
             figsize: The size of the figure
             save: Whether to save the resulting graph to probility.png
+
+        Usage:
+
+        Make sure to fit the model before and only input the
+        probabilities of a single document:
+
+        ```python
+        model.visualize_distribution(probabilities[0])
+        ```
+
+        ![](../img/probabilities.png)
         """
 
         # Get values and indices equal or exceed the minimum probability
@@ -431,12 +667,32 @@ class BERTopic:
             fig.savefig("probability.png", dpi=300, bbox_inches='tight')
 
     def save(self, path: str) -> None:
-        """ Saves the model to the specified path """
+        """ Saves the model to the specified path
+
+        Arguments:
+            path: the location and name of the file you want to save
+
+        Usage:
+
+        ```python
+        model.save("my_model")
+        ```
+        """
         with open(path, 'wb') as file:
             joblib.dump(self, file)
 
     @classmethod
     def load(cls, path: str):
-        """ Loads the model from the specified path """
+        """ Loads the model from the specified path
+
+        Arguments:
+            path: the location and name of the BERTopic file you want to load
+
+        Usage:
+
+        ```python
+        BERTopic.load("my_model")
+        ```
+        """
         with open(path, 'rb') as file:
             return joblib.load(file)
